// Copyright (c) 2017, Herman Bergwerf. All rights reserved.
// Use of this source code is governed by a MIT-style license
// that can be found in the LICENSE file.

part of editex;

class FragmentInsertResult {
  final int cursorOffset;
  final bool didInsert;
  FragmentInsertResult(this.didInsert, [this.cursorOffset = 0]);
}

class FragmentRemoveResult {
  final int count;
  final bool didRemove, removeSelf;
  FragmentRemoveResult(this.didRemove,
      [this.count = 0, this.removeSelf = false]);
  factory FragmentRemoveResult.doRemoveSelf() =>
      new FragmentRemoveResult(false, 0, true);
}

/// Expression fragment.
abstract class Fragment {
  Fragment clone();

  String render(W<int> cursorIdx);

  FragmentInsertResult insert(
          W<int> cursorIdx, Fragment fragment, EdiTeXInterface i) =>
      new FragmentInsertResult(false);

  /// [direction]: -1 for backspace and 0 for delete.
  FragmentRemoveResult remove(W<int> cursorIdx, int direction) =>
      new FragmentRemoveResult(false);

  /// If a command is finished this will return the number of cursor positions
  /// that were part of the command.
  int finish(W<int> cursorIdx, EdiTeXInterface i, bool ifEndsWithDigit) => 0;

  String getParsableContent() => '';

  int get length;
}

/// Fragment containing a single text character.
class CharFragment extends Fragment {
  static const overrideRender = const {r'\': r'\backslash'};

  String char;

  CharFragment(this.char);

  @override
  CharFragment clone() => new CharFragment(char);

  @override
  String render(cursorIdx) {
    if (overrideRender.containsKey(char)) {
      return overrideRender[char];
    }
    return '\\text{$char}';
  }

  @override
  String getParsableContent() => char;

  @override
  int get length => 1;
}

/// Fragment containing a command.
class CommandFragment extends Fragment {
  static final _argRegex = new RegExp(r'\$([0-9]+)');

  final EdiTeXCommand data;
  List<MultiFragment> args;

  CommandFragment(this.data, [this.args]) {
    if (args == null) {
      args = new List<MultiFragment>();

      // Create fragment for each argument.
      data.latexTemplate.replaceAllMapped(_argRegex, (match) {
        args.add(new MultiFragment());
        return '';
      });
    }
  }

  @override
  CommandFragment clone() =>
      new CommandFragment(data, args.map((arg) => arg.clone()).toList());

  @override
  FragmentInsertResult insert(cursorIdx, fragment, interface) {
    for (final arg in args) {
      final result = arg.insert(cursorIdx, fragment, interface);
      if (result.didInsert) {
        return result;
      }
      cursorIdx.v--;
    }
    return new FragmentInsertResult(false);
  }

  @override
  FragmentRemoveResult remove(cursorIdx, direction) {
    for (final arg in args) {
      final result = arg.remove(cursorIdx, direction);
      if (result.didRemove || result.removeSelf) {
        return result;
      }
      cursorIdx.v--;
    }
    return new FragmentRemoveResult(false);
  }

  @override
  int finish(cursorIdx, commands, ifEndsWithDigit) {
    for (final arg in args) {
      final result = arg.finish(cursorIdx, commands, ifEndsWithDigit);
      if (result != 0) {
        return result;
      }
      cursorIdx.v--;
    }
    return 0;
  }

  @override
  String render(cursorIdx) {
    final tex = data.latexTemplate.replaceAllMapped(_argRegex, (match) {
      cursorIdx.v--;
      final idx = int.parse(match.group(1));
      return args[idx].render(cursorIdx);
    });

    // When to add spacing after a command is hard to determine. Since we do not
    // want this to be a burden for the command suplier, we try to establish a
    // relatively common rule here. A space is added when the command ends with
    // a parameter. In the case of \sqrt this means the cursor does not actually
    // move when moving out of the command.
    var appendSpacing = '';
    if (new RegExp(r'\$[0-9]+$').hasMatch(data.latexTemplate)) {
      appendSpacing = '~';
    }

    // Render as block or else spacing will be affected when the cursor is
    // inserted after this (since the cursor is wrapped in a block with the
    // previous element to not affect height).
    if (new RegExp(r'[\^_]').matchAsPrefix(tex) == null) {
      return '{$tex$appendSpacing}';
    } else {
      return '$tex$appendSpacing';
    }
  }

  @override
  String getParsableContent() {
    final str = data.parseTemplate.replaceAllMapped(_argRegex, (match) {
      final idx = int.parse(match.group(1));
      return args[idx].getParsableContent();
    });

    // Add space before and after command to make implicit multiplication
    // obvious.
    return ' $str ';
  }

  @override
  int get length => args.fold(1, (prev, arg) => prev + arg.length);
}

/// Fragment containing multiple child fragments.
class MultiFragment extends Fragment {
  List<Fragment> fragments;

  MultiFragment([this.fragments]) {
    if (fragments == null) {
      fragments = new List<Fragment>();
    }
  }

  @override
  MultiFragment clone() =>
      new MultiFragment(fragments.map((frag) => frag.clone()).toList());

  /// Sequential sub and sup commands should be replaced by a composite command.
  bool _checkSequentialSubSup() {
    if (fragments.isEmpty) {
      return false;
    }

    var prev = fragments.first;
    for (var i = 1; i < fragments.length; i++) {
      final curr = fragments[i];
      if (prev is CommandFragment && curr is CommandFragment) {
        MultiFragment subArg, supArg;
        if (prev.data.name == '_' && curr.data.name == '^') {
          subArg = prev.args.first;
          supArg = curr.args.first;
        } else if (curr.data.name == '_' && prev.data.name == '^') {
          subArg = curr.args.first;
          supArg = prev.args.first;
        }

        if (subArg != null && supArg != null) {
          final command = new CommandFragment(_subSupCommand, [subArg, supArg]);
          fragments.removeRange(i - 1, i + 1);
          fragments.insert(i - 1, command);

          // In theory only one insert or delete happens at a time, so at most
          // one sequential sub/sup can be produced every time.
          return true;
        }
      }

      prev = curr;
    }

    return false;
  }

  @override
  FragmentInsertResult insert(cursorIdx, fragment, interface) {
    for (var i = 0; i <= fragments.length; i++) {
      if (cursorIdx.v == 0) {
        // Special case: forward slash for division, turn into frac command.
        if (fragment is CharFragment && fragment.char == '/') {
          // Get fraction numerator part.
          i--; // Go to previous fragment.
          final startI = i;
          for (; i >= 0; i--) {
            final fragment = fragments[i];
            if (fragment is CommandFragment &&
                (fragment.data.name == '+' || fragment.data.name == '-')) {
              break;
            }
          }

          i++; // -1 => 0 or +/- text fragment => first fragment of numerator.

          final numerator = new MultiFragment();
          numerator.fragments.addAll(fragments.getRange(i, startI + 1));
          fragments.removeRange(i, startI + 1);

          final template = interface.resolveCommand(_divideCommand);
          if (template != null) {
            final command = new CommandFragment(template);
            command.args[0] = numerator;
            fragments.insert(i, command);
          }

          return new FragmentInsertResult(true, 2);
        } else {
          fragments.insert(i, fragment);
          if (_checkSequentialSubSup()) {
            if (fragment is CommandFragment) {
              if (fragment.data.name == '^') {
                return new FragmentInsertResult(true, 0);
              } else if (fragment.data.name == '_') {
                // Select what should be the 'subsup' command.
                // Note that the two sequential commands are turned into one.
                final prev = fragments[i - 1];
                if (prev is CommandFragment && prev.data.name == 'subsup') {
                  return new FragmentInsertResult(true, -prev.args[1].length);
                }
              }
            }
          }

          return new FragmentInsertResult(true, 1);
        }
      } else if (i < fragments.length) {
        cursorIdx.v--;
        final result = fragments[i].insert(cursorIdx, fragment, interface);
        if (result.didInsert) {
          return result;
        }
      }
    }
    return new FragmentInsertResult(false);
  }

  @override
  FragmentRemoveResult remove(cursorIdx, direction) {
    for (var i = 0; i <= fragments.length; i++) {
      if (cursorIdx.v == 0) {
        final idx = i + direction;
        if (idx >= 0 && idx < fragments.length) {
          final removed = fragments.removeAt(i + direction);
          _checkSequentialSubSup();
          return new FragmentRemoveResult(true, removed.length);
        } else {
          return new FragmentRemoveResult.doRemoveSelf();
        }
      } else if (i < fragments.length) {
        final cursorIdxOffsetBefore = cursorIdx.v;
        cursorIdx.v--;
        final result = fragments[i].remove(cursorIdx, direction);

        if (result.removeSelf) {
          fragments.removeAt(i);
          _checkSequentialSubSup();
          return new FragmentRemoveResult(true, cursorIdxOffsetBefore);
        } else if (result.didRemove) {
          return result;
        }
      }
    }
    return new FragmentRemoveResult(false);
  }

  @override
  int finish(cursorIdx, interface, ifEndsWithDigit) {
    for (var i = 0; i < fragments.length; i++) {
      cursorIdx.v--;
      if (cursorIdx.v == 0) {
        if (ifEndsWithDigit) {
          // Only proceed if the character before the cursor is a digit.
          final prevFrag = fragments[i];
          if (prevFrag is CharFragment &&
              !new RegExp(r'^[0-9]$').hasMatch(prevFrag.char)) {
            return 0;
          }
        }

        // Append all text fragments before the cursor.
        final command = new List<int>();
        final startIdx = i;
        var allowNumbers = true;
        for (; i >= 0; i--) {
          final fragi = fragments[i];
          if (fragi is CharFragment) {
            if (fragi.char == r'\') {
              i--; // Ensure removal of backslash.
              break;
            } else if (new RegExp(r'^[0-9]$').hasMatch(fragi.char)) {
              if (!allowNumbers) {
                // Break at numbers.
                break;
              }
            } else {
              allowNumbers = false;
            }
            command.insert(0, fragi.char.codeUnitAt(0));
          } else {
            break;
          }
        }

        final commandStr = new String.fromCharCodes(command);
        final template = interface.resolveCommand(commandStr);
        if (template != null) {
          fragments.removeRange(i + 1, startIdx + 1);
          final fragment = new CommandFragment(template);
          fragments.insert(i + 1, fragment);
          return -(startIdx - i);
        } else {
          return 1;
        }
      } else {
        final result =
            fragments[i].finish(cursorIdx, interface, ifEndsWithDigit);
        if (result != 0) {
          return result;
        }
      }
    }
    return 0;
  }

  @override
  String render(cursorIdx) {
    final buffer = new StringBuffer();
    if (cursorIdx.v == 0) {
      buffer.write(_cursorCommand);
    }
    if (fragments.isNotEmpty) {
      for (var i = 0; i < fragments.length; i++) {
        final fragment = fragments[i];
        final fragTex = fragment.render(cursorIdx);
        cursorIdx.v--;

        if (cursorIdx.v == 0) {
          if (fragment is CommandFragment &&
              new RegExp(r'[\^_]').matchAsPrefix(fragTex) == null) {
            buffer.write('{$fragTex$_cursorCommand}');
          } else {
            buffer.write('$fragTex$_cursorCommand');
          }
        } else {
          buffer.write(fragTex);
        }
      }
    } else {
      buffer.write(_emptyCommand);
    }
    return buffer.toString();
  }

  List getData() {
    final list = new List();

    // Concat sequential char fragments.
    var prevIsChar = false;
    for (final fragment in fragments) {
      if (fragment is CharFragment) {
        if (prevIsChar) {
          // Concatenate
          list[list.length - 1] = '${list.last}${fragment.char}';
        } else {
          list.add(fragment.char);
          prevIsChar = true;
        }
      } else if (fragment is CommandFragment) {
        final commandData = new List()
          ..add(fragment.data.name)
          ..addAll(fragment.args.map((arg) => arg.getData()));
        list.add(commandData);
        prevIsChar = false;
      } else {
        throw new StateError(
            'MultiFragment contains fragments of type ${fragment.runtimeType}');
      }
    }

    return list;
  }

  void loadData(List data, EdiTeXInterface interface) {
    for (final item in data) {
      if (item is String) {
        // Unpack into multpile CharFragments.
        for (final char in item.runes) {
          fragments.add(new CharFragment(new String.fromCharCode(char)));
        }
      } else if (item is List) {
        // Construct [CommandData].
        final command = item.first;
        var commandData = interface.resolveInstantCommand(command);
        if (commandData == null) {
          commandData = interface.resolveCommand(command);
          if (commandData == null) {
            throw new ArgumentError("command '$command' cannot be resolved");
          }
        }

        // Construct arguments.
        final args = new List<MultiFragment>();
        for (var i = 1; i < item.length; i++) {
          args.add(new MultiFragment()..loadData(item[i], interface));
        }

        fragments.add(new CommandFragment(commandData, args));
      }
    }
  }

  @override
  String getParsableContent() =>
      fragments.fold('', (prev, frag) => '$prev${frag.getParsableContent()}');

  @override
  int get length => fragments.fold(1, (prev, arg) => prev + arg.length);
}
